module Gallus
  class Log
    @@global_context_mutex = Mutex.new
    @@global_context = {}

    class << self
      def configure(name = '', &block)
        Repository.get_or_create_logger(name, &block)
      end

      def [](name)
        configure(name)
      end

      def delete(name)
        Repository.delete_with_children(name)
      end

      def define_log_methods!(log_level)
        Level.each do |level|
          method_name = level.name.downcase
          remove_method(method_name) if method_defined?(method_name)

          if level >= log_level
            define_method(method_name) do |message, payload = {}|
              log(level, message, payload)
            end
          else
            define_method(method_name) do |message, payload = {}|
              # supressed...
            end
          end
        end
      end

      def global_context
        @@global_context_mutex.synchronize { yield @@global_context }
      end

      def current_thread_context
        yield (Thread.current[:log_context] ||= {})
      end

      def root
        @@root
      end
    end

    attr_accessor :name, :level, :serialization
    attr_reader :output

    def initialize(parent, name, &block)
      @parent, @name = parent, name.to_s

      if parent
        @output = (@parent.output || []).dup
        @serialization = @parent.serialization
        self.level = @parent.level
      end

      @output ||= []
      @serialization ||= Serialization::Inspect.new

      yield self if block_given?
    end

    def level=(level)
      level = level.is_a?(Level) ? level : Level[level.to_s]
      return if @level == level
      self.class.class_eval { define_log_methods!(level) }
      @level = level
    end

    def context
      @context ||= {}
    end

    private

    def merged_context
      {}.tap do |merged|
        self.class.global_context { |ctx| merged.merge!(ctx) }
        self.class.current_thread_context { |ctx| merged.merge!(ctx) }
        merged.merge!(context)
      end
    end

    def log(level, message, payload = {}, &block)
      payload = merged_context.merge(payload)
      serialized_payload = serialization.call(Payload.new(payload).to_h)
      event = Event.new(@name, level, message.to_s, serialized_payload.empty? ? nil : serialized_payload)
      output.each { |out| out.call(event) }
    end

    @@root = Log.configure do |log|
      log.name = 'root'
      log.level = Level::INFO
      log.serialization = Serialization::Inspect.new
    end
  end
end
