module Perfume
  module Shell
    # Public: Go for Exec when you need to obtain standard output of the command. It uses
    # the %x[] call under the hood. Note that it doesn't raise errors when return status
    # is different than 0. Same as with Perfume::Shell::SystemCall, you should go for
    # inheriting from this class:
    #
    #   class GitDirtyTreeValidation < Perfume::Shell::Exec
    #     Error = Class.new(StandardError)
    #
    #     def cmd
    #       'git status -s'
    #     end
    #
    #     def fail_on_error?
    #       true
    #     end
    #
    #     def handle
    #       raise Error, "Uncommited changes detected!" unless output.strip.empty?
    #     end
    #   end
    #
    class Exec < Base
      args :fail_on_error
      attr_reader :output, :rc

      def init
        @__done = []
        @fail_on_error = false if @fail_on_error.nil?
      end

      def fail_on_error?
        @fail_on_error
      end

      def call(&block)
        Dir.chdir(@root.to_s) do
          before
          log.debug("Executing shell command and waiting for result", cmd: cmd, root: @root.to_s)
          @output, @rc = [ %x[#{cmd} 2>&1], $?.to_i ]
          fail! if fail_on_error? and @rc != 0
          result = respond_to?(:handle) ? handle : [ @output, @rc ]
          yield result if block_given?
          return result
        end
      end

      protected

      def fail!(*)
        $stderr.write(@output)
        super
      end
    end
  end
end
